///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2004 Columbia University.  All Rights Reserved.
//
//      $Header: /Goblin/Patcher/MetadataIterators.h 3     5/28/05 1:20p Eaddy $
//
//    File name:	MetadataIterators.h
//   Created By:	Marc Eaddy
//   Created on:	10/10/04
//
//      Purpose:	STL-like iterators for Unmanaged Metadata
//
//  Description:	Wraps the IMetadataImport->EnumXXX methods in
//					convenient STL iterators.
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Goblin/Patcher/MetadataIterators.h $
 * 
 * 3     5/28/05 1:20p Eaddy
 * Changed copyright
 * Revision 1.2  2004/10/10 06:07:45  me133
 * o Added header comment
 */

#ifndef INCLUDE_METADATA_ITERATORS
#define INCLUDE_METADATA_ITERATORS

#include <cor.h>			// HCORENUM

#include <atlcomcli.h>		// CComPtr

#include <functional>		// STL less<>
#include <iterator>			// STL iterator
#include <string>			// STL string

// MACROS

#define NumItems(s) (sizeof(s) / sizeof(s[0]))

// CONSTANTS

static const int nENUM_BUFFER_SIZE = 50;

//////////////////////
// HCORENUM_iterator
//////////////////////
//
// Base class used by all the iterators
//
template<class Def> 
	class HCORENUM_iterator : 
		public std::iterator<std::input_iterator_tag,
                                            Def>
{
public:

	HCORENUM_iterator(IMetaDataImport* piImport, LPCBYTE pbLoadAddress) 
		: m_spiImport(piImport)
		, m_pbLoadAddress(pbLoadAddress)
		, m_hEnum(0)
		, m_ulCount(0)
		, m_ulTotal(0)
    { 
		_ASSERTE(pbLoadAddress);
	}

	HCORENUM_iterator(IMetaDataImport* piImport) 
		: m_spiImport(piImport)
		, m_pbLoadAddress(0)
		, m_hEnum(0)
		, m_ulCount(0)
		, m_ulCurrent(0)
		, m_ulTotal(0)
    { }

    HCORENUM_iterator(const HCORENUM_iterator& that)
		: m_spiImport(that.m_spiImport)
		, m_pbLoadAddress(that.m_pbLoadAddress)
		, m_hEnum(NULL)
		, m_ulCount(0)
		, m_ulCurrent(0)
		, m_ulTotal(0)
	{ }

	~HCORENUM_iterator()
	{
		if (m_spiImport && m_hEnum)
		{
			m_spiImport->CloseEnum(m_hEnum);
			m_hEnum = NULL;
		}
	}

	// Increments the current element. If we've reached the
	// end of this batch, get the next batch.
    HCORENUM_iterator& operator++()
    {
		if (++m_ulCurrent >= m_ulCount)
		{
			getNextBatch();
		}
		else
		{
			initCurrent();
		}

        return *this;
    }

    const Def& operator*() const
    {
		_ASSERTE(m_ulCurrent < m_ulCount && m_ulCurrent < NumItems(m_ary));
        return m_ary[m_ulCurrent];
    }

    Def operator*()
    {
		_ASSERTE(m_ulCurrent < m_ulCount && m_ulCurrent < NumItems(m_ary));
        return m_ary[m_ulCurrent];
    }

	const Def getToken() const
	{
		return this->operator*();
	}

	bool end() const
	{
		return m_ulCount == 0;
	}

	void reset()
	{
		if (m_spiImport && m_hEnum)
		{
			m_spiImport->CloseEnum(m_hEnum);
			m_hEnum = NULL;
		}

		m_ulCount = 0;
		m_ulCurrent = 0;
		m_ulTotal = 0;
		getNextBatch();
	}

	ULONG count() const
	{
		_ASSERTE(end());
		return m_ulTotal;
	}

	IMetaDataImport* getImport()
	{
		return m_spiImport.p; // Caller must AddRef()/Release()
	}

	LPCBYTE getLoadAddress()
	{
		return m_pbLoadAddress;
	}

protected:
	void getNextBatch()
	{
		// Restart the current position in the array
		m_ulCurrent = 0;

		getNextBatchImpl();

		m_ulTotal += m_ulCount;

		if (!end())
			initCurrent();
	}

	// Init current is called to allow a derived class to initialize any data
	// it may keep that's associated with the current element. When the iterator
	// is incremented, we need to make sure the data is initialized.
	virtual void initCurrent()
	{ }

	virtual void getNextBatchImpl() = 0;

private: // NOT IMPLEMENTED
	HCORENUM_iterator& operator=(const HCORENUM_iterator& that);

    bool operator==(const HCORENUM_iterator& that) const
    {
		return	m_hEnum == that.m_hEnum &&
				m_ulCurrent == that.m_ulCurrent &&
				m_ulCount == that.m_ulCount &&
				m_ulTotal == that.m_ulTotal;
    }

    bool operator!=(const HCORENUM_iterator& that) const
    {
        return !(*this == that);
    }

protected: // DATA
	CComPtr<IMetaDataImport> m_spiImport;
	LPCBYTE m_pbLoadAddress;
    HCORENUM m_hEnum;
    Def m_ary[nENUM_BUFFER_SIZE];
	ULONG m_ulCount;

private:	// DATA
	ULONG m_ulCurrent;
	ULONG m_ulTotal;
};

/////////////////////
// typeDef_iterator
/////////////////////

class typeDef_iterator : public HCORENUM_iterator<mdTypeDef>
{
public:
	typedef HCORENUM_iterator<mdTypeDef> _Mybase;

	typeDef_iterator(IMetaDataImport* piImport, LPCBYTE pbLoadAddress);

	bool isInterface() const;
	bool isDelegate() const;
	const WCHAR* getName() const;

private: // HELPER FUNCTIONS
	void getNextBatchImpl();
	void initCurrent();
	HRESULT getProps() const;
	HRESULT getProps(mdToken tokType, WCHAR* szTypeName, ULONG nMaxLen, DWORD* pdwFlags, mdToken* ptokBaseType) const;

private: // NOT IMPLEMENTED
    typeDef_iterator(const typeDef_iterator& it); 
	typeDef_iterator& operator=(const typeDef_iterator& that);

private: // DATA
	mutable WCHAR m_szTypeName[2048]; // 2048 from alink.h
	mutable DWORD m_dwFlags;
	mutable mdToken m_tokBaseType;
};

///////////////////////
// methodDef_iterator
///////////////////////

class methodDef_iterator : public HCORENUM_iterator<mdMethodDef>
{
public:
	typedef HCORENUM_iterator<mdMethodDef> _Mybase;

    methodDef_iterator(IMetaDataImport* piImport, LPCBYTE pbLoadAddress, mdTypeDef typeDef);
	methodDef_iterator(typeDef_iterator& itTypeDef);

private: // HELPER FUNCTIONS
	void getNextBatchImpl();

private: // NOT IMPLEMENTED
    methodDef_iterator(const methodDef_iterator& it); 
	methodDef_iterator& operator=(const methodDef_iterator& that);

private: // DATA
	mdTypeDef m_typeDef;
};

//////////////////////
// paramDef_iterator
//////////////////////

class paramDef_iterator : public HCORENUM_iterator<mdParamDef>
{
public:
	typedef HCORENUM_iterator<mdParamDef> _Mybase;

    paramDef_iterator(IMetaDataImport* piImport, LPCBYTE pbLoadAddress, mdMethodDef methodDef);
	paramDef_iterator(methodDef_iterator& itMethodDef);

private: // HELPER FUNCTIONS
	void getNextBatchImpl();

private: // NOT IMPLEMENTED
    paramDef_iterator(const paramDef_iterator& it); 
	paramDef_iterator& operator=(const paramDef_iterator& that);

private: // DATA
	mdMethodDef m_methodDef;
};

struct SStrWLen
{
	std::wstring s;
	ULONG len;

	friend bool operator<(const SStrWLen& lhs, const SStrWLen& rhs)
	{
		int nCmp = lhs.s.compare(rhs.s);
		return nCmp < 0 || (nCmp == 0 && lhs.len < rhs.len);
	}
};

// This method is needed by map<SStrWLen, mdToken> to properly sort
bool std::less<SStrWLen>::operator()(const SStrWLen& lhs, const SStrWLen& rhs) const
{
	return lhs < rhs;
}

////////////////////////
// userString_iterator
////////////////////////

class userString_iterator : public HCORENUM_iterator<mdString>
{
public:
	typedef HCORENUM_iterator<mdString> _Mybase;

    userString_iterator(IMetaDataImport* piImport);
	~userString_iterator();

	// The string returned will be invalid if you increment or
	// destroy the iterator
	const SStrWLen& getStr() const;

private: // HELPER FUNCTIONS
	void getNextBatchImpl();
	void reserve(ULONG nCapacity) const;

private: // NOT IMPLEMENTED
    userString_iterator(const userString_iterator& it); 
	userString_iterator& operator=(const userString_iterator& that);

private:
	mutable WCHAR* m_pszTemp;
	mutable ULONG  m_ulSize;

	static const ULONG nINITIAL_SIZE	= 50;
	static const ULONG nINC_FACTOR		= 2;
};

//////////////////////
// signature_iterator
//////////////////////

class signature_iterator : public HCORENUM_iterator<mdSignature>
{
public:
	typedef HCORENUM_iterator<mdSignature> _Mybase;

    signature_iterator(IMetaDataImport* piImport);

	PCCOR_SIGNATURE getSignature() const;

private: // HELPER FUNCTIONS
	void getNextBatchImpl();

private: // NOT IMPLEMENTED
    signature_iterator(const signature_iterator& it); 
	signature_iterator& operator=(const signature_iterator& that);
};

#endif // INCLUDE_METADATA_ITERATORS
