///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2004 Columbia University.  All Rights Reserved.
//
//      $Header: /Goblin/Watcher/AspectRules.cs 8     5/28/05 1:24p Eaddy $
//
//    File name:	AspectRules.cs
//   Created By:	Marc Eaddy
//   Created on:	10/17/04
//
//      Purpose:	
//
//  Description:	
//
// Currently, the C++/CLI and J# compilers do not work.  The C++/CLI CodeDomProvider
// returns null when CreateCompiler() is called.  The J# compiler compiles but there
// are problems when the function is patched that cause an InvalidProgramException.
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Goblin/Watcher/AspectRules.cs $
 * 
 * 8     5/28/05 1:24p Eaddy
 * o Changed copyright
 * 
 * 7     5/25/05 6:33p Eaddy
 * o Now uses EnC.Watcher namespace
 * 
 * 6     5/10/05 5:10p Eaddy
 * o Now fails gracefully for field decls
 * 
 * 5     4/22/05 7:36p Eaddy
 * o Can now turn off weaving using an attribute in the XML file
 * 
 * 4     4/15/05 2:49p Eaddy
 * o Added assert
 * 
 * 3     4/04/05 2:28p Eaddy
 * o Supports Properties
 * o Added AspectRulesFile class
 * 
 * 2     4/01/05 6:56p Eaddy
 * o Now handles parsing exceptions
 * 
 * 1     4/01/05 11:15a Eaddy
 * o Now supports rudimentary weaving
 */

using System;
using System.IO;
using System.Reflection;
using System.Diagnostics;
using System.Collections;
using System.Xml;
using System.Xml.Serialization;

using DDW.CSharp.Dom;

namespace EnC.Watcher
{
	public class Pointcut
	{
		[XmlAttributeAttribute("element")]
		public string codeElementTypeName;
		
		[XmlAttributeAttribute("name")]
		public string elementPattern;

		public GraphTypes codeElementGraphType = (GraphTypes)(-1);

		public void init()
		{
			if (codeElementTypeName == null || codeElementTypeName.Length == 0)
			{
				return;
			}

			if (codeElementTypeName.IndexOf('.') == -1)
			{
				codeElementTypeName = "DDW.CSharp.Dom." + codeElementTypeName + "Decl";
			}

			Type type = Helper.GetType(codeElementTypeName);
			if (type == null)
			{
				MyTrace.Log(TraceLevel.Error, "Failed to load type '{0}'.", codeElementTypeName);
				return;
			}

			try
			{
				IDeclaration decl = (IDeclaration) Activator.CreateInstance(type);
				if (decl == null)
				{
					MyTrace.Log(TraceLevel.Error, "Failed to create instance of type '{0}'.", codeElementTypeName);
					return;
				}

				codeElementGraphType = decl.GraphType;
			}
			catch (TypeInitializationException ex)
			{
				MyTrace.Log(TraceLevel.Error, "Failed to create instance of type '{0}': {1}", 
					codeElementTypeName, ex.Message);
			}
			catch (InvalidCastException ex)
			{
				MyTrace.Log(TraceLevel.Error, "Failed to create instance of type '{0}': {1}", 
					codeElementTypeName, ex.Message);
			}
		}
	}

	public class IGraphMatch
	{
		public IGraphMatch(IGraph parent, IGraph match)
		{
			this.match = match;
			this.parent = parent;
		}

		public IGraph match;
		public IGraph parent;
	}

	public class AspectRule
	{
		[XmlAttributeAttribute("namespace")]
		public string namespaceName;

		[XmlAttributeAttribute("class")]
		public string className;

		[XmlAttributeAttribute("sources")]
		public string sources;

		[XmlElement("pointcut")]
		public Pointcut[] pointcuts = null;

		IGraph ig = null;

		MethodDecl before = null;
		MethodDecl after = null;
		MethodDecl around = null;

		static Type typeCEA = typeof(DDW.CSharp.Dom.CodeElementAttribute);

		public void init()
		{
			Debug.Assert(sources != null);

			foreach(Pointcut pointcut in pointcuts)
			{
				pointcut.init();
			}

			try
			{
				ig = Helper.GetGraph(sources);
				if (ig == null)
				{
					return;
				}
			}
			catch(Exception ex)
			{
				MyTrace.Log(TraceLevel.Error, "Failed to parse source file '{0}': {1}",
					sources, ex.ToString());
				return;
			}

			before = findMethod(ig, "before");
			after = findMethod(ig, "after");
			around = findMethod(ig, "around");

			Debug.Assert(before != null || after != null || around != null);
			Debug.Assert(around == null || (before == null && after == null));
		}

		public int instrument(IGraph ig)
		{
			ArrayList matches = new ArrayList();

			foreach(Pointcut pointcut in pointcuts)
			{
				findElements(null, ig, pointcut.codeElementGraphType, pointcut.elementPattern, matches);
			}

			// Keeps track of the number of instrumented code elements
			int countInstrumented = 0;

			for(int i = 0; i < matches.Count; ++i)
			{
				IGraph igMatch = ((IGraphMatch) matches[i]).match;

				bool instrBefore = instrumentBefore(igMatch);
				bool instrAfter  = instrumentAfter(igMatch);
				bool instrAround = instrumentAround(igMatch);

				if (instrBefore || instrAfter || instrAround)
					++countInstrumented;
			}

			return countInstrumented;
		}

		public bool instrumentBefore(IGraph ig)
		{
			if (before == null)
				return false;

			if (ig.GraphType == GraphTypes.MethodDecl)
			{
				MethodDecl declMethod = (MethodDecl) ig;
				return instrumentBefore(declMethod.Statements);
			}
			else if (ig.GraphType == GraphTypes.PropertyDecl)
			{
				PropertyDecl declProp = (PropertyDecl) ig;
				instrumentBefore(declProp.GetAccessor.Statements);
				return instrumentBefore(declProp.SetAccessor.Statements);
			}

			return false;
		}

		private bool instrumentBefore(StatementCollection statements)
		{
			if (before == null)
				return false;

			for(int i = 0; i < before.Statements.Count; ++i)
			{
				statements.Insert(i, before.Statements[i]);
			}

			return before.Statements.Count > 0;
		}

		public bool instrumentAfter(IGraph ig)
		{
			if (after == null)
				return false;

			if (ig.GraphType == GraphTypes.MethodDecl)
			{
				MethodDecl declMethod = (MethodDecl) ig;
				return instrumentAfter(declMethod.Statements);
			}
			else if (ig.GraphType == GraphTypes.PropertyDecl)
			{
				PropertyDecl declProp = (PropertyDecl) ig;

				instrumentAfter(declProp.GetAccessor.Statements);
				return instrumentAfter(declProp.SetAccessor.Statements);
			}

			return false;
		}

		private bool instrumentAfter(StatementCollection body)
		{
			if (after == null)
				return false;

			const bool useTryFinally = true;

			if (useTryFinally)
			{
				StatementCollection bodyCopy = new StatementCollection(body);
				TryCatchFinallyStmt tryCatchFinallyStmt = new TryCatchFinallyStmt();
				tryCatchFinallyStmt.Try.AddRange(bodyCopy);
				tryCatchFinallyStmt.Finally.AddRange(after.Statements);
				body.Clear();
				body.Add(tryCatchFinallyStmt);
			}
	/*		else
			{
				// Search for all the return statements

				ArrayList matches = new ArrayList();
				findElements(null, body, GraphTypes.ReturnStmt, null, matches);

				for(int i = 0; i < matches.Count; ++i)
				{
					IGraphMatch matchReturn = (IGraphMatch) matches[i];
					
					StatementCollection statements = matchReturn.parent as StatementCollection;
					if (statements == null)
						continue; // We only know how to instrument statement collections

					// Locate the matched return statement. The remaining return statements
					// will be instrumented in subsequent iterations. This handles multiple return
					// statements in the same StatementCollection as well as return statements in
					// other StatementCollections.
					int indexRetStmt = statements.IndexOf( (Statement) matchReturn.match);
					Debug.Assert(indexRetStmt >= 0);

					// Insert the after aspect statements before the return statement
					for(int k = 0; k < after.Statements.Count; ++k)
					{
						statements.Insert(indexRetStmt+k, after.Statements[k]);
					}
				}

				int count = body.Count;
				if (count > 0 && body[count-1].GraphType != GraphTypes.ReturnStmt)
				{
					body.AddRange(after.Statements);
				}
			}
	*/
			return true;
		}

		public bool instrumentAround(IGraph ig)
		{
			if (around == null)
				return false;

			if (ig.GraphType == GraphTypes.MethodDecl)
			{
				MethodDecl declMethod = (MethodDecl) ig;

				declMethod.Statements.Clear();
				declMethod.Statements.AddRange(around.Statements);

				return true;
			}
			else if (ig.GraphType == GraphTypes.PropertyDecl)
			{
				PropertyDecl declProp = (PropertyDecl) ig;

				declProp.GetAccessor.Statements.Clear();
				declProp.GetAccessor.Statements.AddRange(after.Statements);
				declProp.SetAccessor.Statements.Clear();
				declProp.SetAccessor.Statements.AddRange(after.Statements);

				return true;
			}

			return false;
		}

		public static int findElements(IGraph parent, IGraph element, GraphTypes elemType, string pattern, ArrayList matches)
		{
			if (elemType == (GraphTypes)(-1) || element.GraphType == elemType)
			{
				if (pattern == null || 
					pattern.Length == 0 ||
					Helper.ReflectGetProperty(element, "Name") == pattern)
				{
					matches.Add(new IGraphMatch(parent, element));
				}
			}

			if (element is ICollection)
			{
				ICollection children = (ICollection) element;
				foreach(IGraph child in children)
				{
					findElements(element, child, elemType, pattern, matches);
				}
			}

			Type type = element.GetType();

			MemberInfo[] props = type.FindMembers(MemberTypes.Property,
				BindingFlags.Public | BindingFlags.Instance,
				new MemberFilter(MemberFilter),
				null);

			foreach(MemberInfo prop in props)
			{
				object codeElementProp = type.InvokeMember(	prop.Name, 
					BindingFlags.Public | BindingFlags.GetProperty |  BindingFlags.Instance, 
					null, 
					element, 
					new object[]{} );

				if (codeElementProp is ICollection)
				{
					ICollection children = (ICollection) codeElementProp;
					parent = (IGraph) codeElementProp;
					foreach(IGraph child in children)
					{
						findElements(parent, child, elemType, pattern, matches);
					}
				}
				else if (codeElementProp is IGraph)
				{
					findElements(element, (IGraph) codeElementProp, elemType, pattern, matches);
				}
			}

			return matches.Count;
		}

		public static MethodDecl findMethod(IGraph element, string methodName)
		{
			ArrayList matches = new ArrayList();
			findElements(null, element, GraphTypes.MethodDecl, methodName, matches);
			if (matches.Count > 0)
			{
				// Just grab the first one
				return (MethodDecl) ((IGraphMatch) matches[0]).match;
			}
			else
			{
				return null;
			}
		}

		public static bool MemberFilter(MemberInfo objMemberInfo, Object objSearch)
		{			
			if(objMemberInfo.Name == "LinePragma") 
				return false;
			
			return objMemberInfo.IsDefined(typeCEA, true);
		}
	}

	[XmlRoot("aspects")]
	public class AspectRules
	{
		[XmlAttributeAttribute("enable")]
		public bool enableWeaving = true;

		[XmlElement("aspect")]
		public AspectRule[] rules = null;

		static public AspectRules load(string path)
		{
			try
			{
				TextReader reader =	new	StreamReader(path);
				XmlSerializer serializer = new XmlSerializer(typeof(AspectRules));
				AspectRules aspectRules = (AspectRules) serializer.Deserialize(reader);
				reader.Close();
				return aspectRules;
			}
			catch(XmlException ex)
			{
				MyTrace.Log(TraceLevel.Error, "Failed to parse XML Aspects file: {0}.",
					ex.Message);
				return null;
			}
			catch(InvalidOperationException	ex)
			{
				MyTrace.Log(TraceLevel.Error, "Failed to parse XML Aspects file: {0}.",
					ex.InnerException.Message);
				return null;
			}
			catch(Exception ex)
			{
				MyTrace.Log(TraceLevel.Error, "Failed to parse XML Aspects file: {0}",
					ex.Message);
				return null;
			}
		}

		public void init()
		{
			foreach(AspectRule rule in rules)
			{
				rule.init();
			}
		}

		public void instrument(IGraph ig)
		{
			if (ig == null)
				return;

			foreach(AspectRule rule in rules)
			{
				rule.instrument(ig);
			}
		}
	}

	public class AspectRulesFile : FileWatcherCollection
	{
		/// <summary>
		/// Delegate that we call when an aspect file has been changed.
		/// </summary>
		event MyFileWatcher.OnChangedDelegate _onFileChangedEvent_ForClients;

		/// <summary>
		/// Delegate that MyFileWatcher calls when a file has changed. We cache this for efficiency.
		/// </summary>
		readonly MyFileWatcher.OnChangedDelegate _onFileChangedDelegate_ForMyself;

		/// <summary>
		/// Aspects
		/// </summary>
		AspectRules				_aspectRules = null;

		/// <summary>
		/// Aspect rule filename.
		/// </summary>
		const string			_aspectRuleFile = "AspectRules.xml";

		public AspectRulesFile()
		{
			// Initialize the file watcher changed delegate
			_onFileChangedDelegate_ForMyself = new MyFileWatcher.OnChangedDelegate(OnFileChanged);

			// Load the aspect rules
			loadAspectRules(_aspectRuleFile);

			// Watch the aspect rule file for changes
			if (_aspectRules != null)
				registerForChanges(_aspectRuleFile);
		}

		/// <summary>
		/// Loads the aspect rule file.
		/// </summary>
		/// <param name="pathAspectRules">Path to aspect rule file</param>
		private void loadAspectRules(string pathAspectRules)
		{
			// Step 1:	Stop watching the old aspect source files

			if (_aspectRules != null)
			{
				foreach(AspectRule rule in _aspectRules.rules)
				{
					unregisterForChanges(rule.sources);
				}
			}

			// Step 2:	Load the new aspect rules

			_aspectRules = AspectRules.load(pathAspectRules);
			if (_aspectRules == null)
			{
				MyTrace.Log(TraceLevel.Error, "Failed to load aspect rules.");
				return;
			}

			// Step 3:	Initialize aspect rules
			//
			//			This causes the aspect source files to be parsed.

			_aspectRules.init();

			// Step 4:	Watch for changes to the aspect source files

			foreach(AspectRule rule in _aspectRules.rules)
			{
				Debug.Assert(rule.sources != null);
				registerForChanges(rule.sources);
			}
		}

		/// <summary>
		/// <seealso cref="AspectRules.instrument"/>
		/// </summary>
		/// <param name="ig"></param>
		public void instrument(IGraph ig)
		{
			if (_aspectRules != null)
			{
				lock(_aspectRules)
				{
					_aspectRules.instrument(ig);
				}
			}
		}

		/// <summary>
		/// Sets the callback used to notify when a module has been recompiled.
		/// </summary>
		/// <param name="onFileChangedDelegate_ForClients"></param>
		public void setFileChangedDelegate(MyFileWatcher.OnChangedDelegate onFileChangedDelegate_ForClients)	
		{
			lock(this)
			{
				_onFileChangedEvent_ForClients += onFileChangedDelegate_ForClients;
			}
		}

		/// <summary>
		/// Internal callback called by MyFileWatcher when a file is changed.
		/// </summary>
		/// <param name="path">Path of changed file.</param>
		protected override void OnFileChanged(string path)
		{
			MyTrace.Log(TraceLevel.Info, "Change detected: " + path + ". Reloading aspects.");

			// If the aspect rule file changed or one of the aspect source files then
			// re-load the aspect rules and re-parse the source files.
			lock(_aspectRuleFile)
			{
				loadAspectRules(_aspectRuleFile);
			}

			// Notify interested listeners

			if (_onFileChangedEvent_ForClients != null)
			{
				lock(_onFileChangedEvent_ForClients)
				{
					_onFileChangedEvent_ForClients(path);
				}
			}
		}

		public bool Enabled
		{
			get { return _aspectRules != null && _aspectRules.enableWeaving; }
		}
	}
}
